@@ -5,7 +5,8 @@
 //! also exported as a library too, so that projects
 //! such as `rpm-ostree` can directly reuse it.
 
-use anyhow::Result;
+use anyhow::{Context, Result};
+use camino::Utf8PathBuf;
 use futures_util::FutureExt;
 use ostree::{cap_std, gio, glib};
 use std::borrow::Borrow;
@@ -17,19 +18,23 @@ use structopt::StructOpt;
 use tokio_stream::StreamExt;
 
 use crate::commit::container_commit;
-use crate::container::store::{LayeredImageImporter, PrepareResult};
-use crate::container::{self as ostree_container, UnencapsulationProgress};
+use crate::container as ostree_container;
 use crate::container::{Config, ImageReference, OstreeImageReference, UnencapsulateOptions};
+use ostree_container::store::{ImageImporter, PrepareResult};
+use ostree_container::UnencapsulationProgress;
 
-fn parse_imgref(s: &str) -> Result<OstreeImageReference> {
+/// Parse an [`OstreeImageReference`] from a CLI arguemnt.
+pub fn parse_imgref(s: &str) -> Result<OstreeImageReference> {
     OstreeImageReference::try_from(s)
 }
 
-fn parse_base_imgref(s: &str) -> Result<ImageReference> {
+/// Parse a base [`ImageReference`] from a CLI arguemnt.
+pub fn parse_base_imgref(s: &str) -> Result<ImageReference> {
     ImageReference::try_from(s)
 }
 
-fn parse_repo(s: &str) -> Result<ostree::Repo> {
+/// Parse an [`ostree::Repo`] from a CLI arguemnt.
+pub fn parse_repo(s: &str) -> Result<ostree::Repo> {
     let repofd = cap_std::fs::Dir::open_ambient_dir(s, cap_std::ambient_authority())?;
     Ok(ostree::Repo::open_at_dir(&repofd, ".")?)
 }
@@ -132,6 +137,8 @@ enum ContainerOpts {
     },
 
     #[structopt(alias = "commit")]
+    /// Perform build-time checking and canonicalization.
+    /// This is presently an optional command, but may become required in the future.
     Commit,
 
     /// Commands for working with (possibly layered, non-encapsulated) container images.
@@ -184,6 +191,17 @@ enum ContainerImageOpts {
         proxyopts: ContainerProxyOpts,
     },
 
+    /// Pull (or update) a container image.
+    History {
+        /// Path to the repository
+        #[structopt(long, parse(try_from_str = parse_repo))]
+        repo: ostree::Repo,
+
+        /// Image reference, e.g. ostree-remote-image:someremote:registry:quay.io/exampleos/exampleos:latest
+        #[structopt(parse(try_from_str = parse_imgref))]
+        imgref: OstreeImageReference,
+    },
+
     /// Copy a pulled container image from one repo to another.
     Copy {
         /// Path to the source repository
@@ -230,6 +248,10 @@ enum ContainerImageOpts {
         #[structopt(long)]
         /// Add a kernel argument
         karg: Option<Vec<String>>,
+
+        /// Write the deployed checksum to this file
+        #[structopt(long)]
+        write_commitid_to: Option<Utf8PathBuf>,
     },
 }
 
@@ -254,7 +276,7 @@ struct ImaSignOpts {
 /// Options for internal testing
 #[derive(Debug, StructOpt)]
 enum TestingOpts {
-    // Detect the current environment
+    /// Detect the current environment
     DetectEnv,
     /// Execute integration tests, assuming mutable environment
     Run,
@@ -273,6 +295,7 @@ enum Opt {
     Container(ContainerOpts),
     /// IMA signatures
     ImaSign(ImaSignOpts),
+    /// Internal integration testing helpers.
     #[structopt(setting(structopt::clap::AppSettings::Hidden))]
     #[cfg(feature = "internal-testing-api")]
     InternalOnlyForTesting(TestingOpts),
@@ -335,16 +358,14 @@ async fn container_import(
     let (tx_progress, rx_progress) = tokio::sync::watch::channel(Default::default());
     let target = indicatif::ProgressDrawTarget::stdout();
     let style = indicatif::ProgressStyle::default_bar();
-    let pb = if !quiet {
+    let pb = (!quiet).then(|| {
         let pb = indicatif::ProgressBar::new_spinner();
         pb.set_draw_target(target);
         pb.set_style(style.template("{spinner} {prefix} {msg}"));
         pb.enable_steady_tick(200);
         pb.set_message("Downloading...");
-        Some(pb)
-    } else {
-        None
-    };
+        pb
+    });
     let opts = UnencapsulateOptions {
         progress: Some(tx_progress),
     };
@@ -410,7 +431,8 @@ async fn container_export(
         copy_meta_keys,
         ..Default::default()
     };
-    let pushed = crate::container::encapsulate(repo, rev, &config, Some(opts), imgref).await?;
+    let pushed =
+        crate::container::encapsulate(repo, rev, &config, Some(opts), None, imgref).await?;
     println!("{}", pushed);
     Ok(())
 }
@@ -428,7 +450,7 @@ async fn container_store(
     imgref: &OstreeImageReference,
     proxyopts: ContainerProxyOpts,
 ) -> Result<()> {
-    let mut imp = LayeredImageImporter::new(repo, imgref, proxyopts.into()).await?;
+    let mut imp = ImageImporter::new(repo, imgref, proxyopts.into()).await?;
     let prep = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(c) => {
             println!("No changes in {} => {}", imgref, c.merge_commit);
@@ -436,17 +458,7 @@ async fn container_store(
         }
         PrepareResult::Ready(r) => r,
     };
-    if prep.base_layer.commit.is_none() {
-        let size = crate::glib::format_size(prep.base_layer.size());
-        println!(
-            "Downloading base layer: {} ({})",
-            prep.base_layer.digest(),
-            size
-        );
-    } else {
-        println!("Using base: {}", prep.base_layer.digest());
-    }
-    for layer in prep.layers.iter() {
+    for layer in prep.all_layers() {
         if layer.commit.is_some() {
             println!("Using layer: {}", layer.digest());
         } else {
@@ -472,6 +484,61 @@ async fn container_store(
     Ok(())
 }
 
+fn print_column(s: &str, clen: usize, remaining: &mut usize) {
+    let l = s.len().min(*remaining);
+    print!("{}", &s[0..l]);
+    if clen > 0 {
+        // We always want two trailing spaces
+        let pad = clen.saturating_sub(l) + 2;
+        for _ in 0..pad {
+            print!(" ");
+        }
+        *remaining = remaining.checked_sub(l + pad).unwrap();
+    }
+}
+
+/// Output the container image history
+async fn container_history(repo: &ostree::Repo, imgref: &OstreeImageReference) -> Result<()> {
+    let img = crate::container::store::query_image(repo, imgref)?
+        .ok_or_else(|| anyhow::anyhow!("No such image: {}", imgref))?;
+    let columns = [("ID", 20), ("SIZE", 10), ("CREATED BY", 0usize)];
+    let width = term_size::dimensions().map(|x| x.0).unwrap_or(80);
+    if let Some(config) = img.configuration.as_ref() {
+        {
+            let mut remaining = width;
+            for (name, width) in columns.iter() {
+                print_column(name, *width as usize, &mut remaining);
+            }
+            println!();
+        }
+
+        let mut history = config.history().iter();
+        let layers = img.manifest.layers().iter();
+        for layer in layers {
+            let histent = history.next();
+            let created_by = histent
+                .and_then(|s| s.created_by().as_deref())
+                .unwrap_or("");
+
+            let mut remaining = width;
+
+            let digest = layer.digest().as_str();
+            // Verify it's OK to slice, this should all be ASCII
+            assert!(digest.chars().all(|c| c.is_ascii()));
+            let digest_max = columns[0].1;
+            let digest = &digest[0..digest_max];
+            print_column(digest, digest_max, &mut remaining);
+            let size = glib::format_size(layer.size() as u64);
+            print_column(size.as_str(), columns[1].1, &mut remaining);
+            print_column(created_by, columns[2].1, &mut remaining);
+            println!();
+        }
+        Ok(())
+    } else {
+        anyhow::bail!("v0 image does not have fetched configuration");
+    }
+}
+
 /// Add IMA signatures to an ostree commit, generating a new commit.
 fn ima_sign(cmdopts: &ImaSignOpts) -> Result<()> {
     let signopts = crate::ima::ImaOpts {
@@ -555,6 +622,9 @@ where
                     imgref,
                     proxyopts,
                 } => container_store(&repo, &imgref, proxyopts).await,
+                ContainerImageOpts::History { repo, imgref } => {
+                    container_history(&repo, &imgref).await
+                }
                 ContainerImageOpts::Copy {
                     src_repo,
                     dest_repo,
@@ -567,6 +637,7 @@ where
                     target_imgref,
                     karg,
                     proxyopts,
+                    write_commitid_to,
                 } => {
                     let sysroot = &ostree::Sysroot::new(Some(&gio::File::for_path(&sysroot)));
                     sysroot.load(gio::NONE_CANCELLABLE)?;
@@ -580,8 +651,18 @@ where
                         target_imgref: target_imgref.as_ref(),
                         proxy_cfg: Some(proxyopts.into()),
                     };
-                    crate::container::deploy::deploy(sysroot, &stateroot, &imgref, Some(options))
-                        .await
+                    let state = crate::container::deploy::deploy(
+                        sysroot,
+                        &stateroot,
+                        &imgref,
+                        Some(options),
+                    )
+                    .await?;
+                    if let Some(p) = write_commitid_to {
+                        std::fs::write(&p, state.merge_commit.as_bytes())
+                            .with_context(|| format!("Failed to write commitid to {}", p))?;
+                    }
+                    Ok(())
                 }
             },
         },
