@@ -32,6 +32,8 @@ use std::ops::Deref;
 
 /// The label injected into a container image that contains the ostree commit SHA-256.
 pub const OSTREE_COMMIT_LABEL: &str = "ostree.commit";
+/// The label/annotation which contains the sha256 of the final commit.
+const OSTREE_DIFFID_LABEL: &str = "ostree.diffid";
 
 /// Our generic catchall fatal error, expected to be converted
 /// to a string to output to a terminal or logs.
@@ -150,7 +152,7 @@ impl TryFrom<&str> for OstreeImageReference {
             // Shorthand for ostree-unverified-image:registry:
             "ostree-unverified-registry" => (
                 SignatureSource::ContainerPolicyAllowInsecure,
-                Cow::Owned(format!("registry:{}", second)),
+                Cow::Owned(format!("registry:{second}")),
             ),
             // This is a shorthand for ostree-remote-image with registry:
             "ostree-remote-registry" => {
@@ -159,7 +161,7 @@ impl TryFrom<&str> for OstreeImageReference {
                     .ok_or_else(|| anyhow!("Missing second ':' in {}", value))?;
                 (
                     SignatureSource::OstreeRemote(remote.to_string()),
-                    Cow::Owned(format!("registry:{}", rest)),
+                    Cow::Owned(format!("registry:{rest}")),
                 )
             }
             "ostree-remote-image" => {
@@ -229,6 +231,15 @@ mod encapsulate;
 pub use encapsulate::*;
 mod unencapsulate;
 pub use unencapsulate::*;
+// We have this trick of compiling ourself with integration testing
+// enabled, which uses a lot of the code here.   See the
+// `ostree-ext = { path = ".", features = ["internal-testing-api"] }`
+// bit in Cargo.toml.
+//
+// But that isn't turned on for other crates that use this, and correctly gating all
+// of it is a little tedious.  So let's just use the big hammer for now to
+// quiet the dead code warnings.
+#[allow(dead_code)]
 pub(crate) mod ocidir;
 mod skopeo;
 pub mod store;
