@@ -26,6 +26,7 @@ type Result<T> = anyhow::Result<T>;
 // Import global functions.
 mod globals;
 
+pub mod bootabletree;
 pub mod cli;
 pub mod container;
 pub mod container_utils;
@@ -36,13 +37,18 @@ pub mod refescape;
 pub mod tar;
 pub mod tokio_util;
 
+pub mod chunking;
 pub(crate) mod commit;
+pub mod objectsource;
 pub(crate) mod objgv;
+
 /// Prelude, intended for glob import.
 pub mod prelude {
     #[doc(hidden)]
     pub use ostree::prelude::*;
 }
 
+#[cfg(feature = "internal-testing-api")]
+pub mod fixture;
 #[cfg(feature = "internal-testing-api")]
 pub mod integrationtest;
